% A three-dimensional cartesian grid is made and this is extruded to three 
% dimensions. This example is made to compare the solutions obtained by
% TPFA, MPFA, Mimetic using inner product 'ip_simple', the hybrid mimetic
% method and the hybrid MPFA method to the analytic solution.


% Making a 10x10x5 cartesian grid where each cell has dimension 1x1x1.
nh = 10; nw = 10; zval = 5;
G = cartGrid([nh nw zval]);

% We continue to compute centroids and volumes of the cells and centroids, 
% normals, and areas for the faces. 
G = computeGeometry(G);


%% Set rock and fluid data

% We set the permeability, rock.perm, to be homogeneous and isotropic 
% K = diag(1,1,1) Darcy. 
rock.perm = repmat([1000, 1000, 1000].* milli*darcy(), [G.cells.num, 1]);
% The viscosity is specified by saying that the reservoir is filled with a 
% single fluid, where the viscosity value is 1 cp. The density of the fluid 
% is set to be 1000 kg/m^3.
fluid     = initSingleFluid('mu' ,    1*centi*poise     , ...
                            'rho', 1000*kilogram/meter^3);

%% Initialize reservoir simulator

% We set the initial reservoir pressure to be 0. The single-phase
% saturation is also set to 0, since this is the default behaviour of
% initResSol.
resSol = initResSol(G, 0.0);
display(resSol);

%% Impose Dirichlet boundary conditions
westFaces = find(G.faces.centroids(:,1) == 0);
bc = addBC([], westFaces, 'pressure', 0);
eastFaces = find(G.faces.centroids(:,1) == nh);
bc = addBC(bc, eastFaces, 'pressure', 100*barsa);   
display(bc);

%% Construct linear system
gravity off;
% TPFA
rt = computeTrans(G, rock);
% Mimetic
rs = computeMimeticIP(G, rock, 'InnerProduct', 'ip_simple');
% Hybrid Mimetic
rh = computeHybridMimeticIP(G, rock, 'InnerProduct', 'ip_hyb');

% Adding the module where the MPFA function is.
mrstModule add mpfa

% MPFA
rmp = computeMultiPointTrans(G, rock);
% Hybrid MPFA
rhmp = computeHybridMultiPointTrans(G, rock);

%% Solve the linear system
resSol = initResSol(G, 0.0);
% solve TPFA system
resSol1 = incompTPFA(resSol, G, rt, fluid, 'bc', bc);
% solve Mimetic system
resSol2 = solveIncompFlow(resSol, G, rs, fluid, ...
                          'bc', bc, 'MatrixOutput', true);
% solve hybrid mimetic system 
resSol3 = solveIncompFlow(resSol, G, rh, fluid, ...
                          'bc', bc, 'MatrixOutput', true);
% solve MPFA system
resSol4 = incompMPFA(resSol, G, rmp, fluid, 'bc', bc);
% solve hybrid MPFA system
resSol5 = incompMPFA(resSol, G, rhmp, fluid, 'bc', bc);


% Computing the analytic pressure solution. Valid in current form when 
% rock.perm = diag(1,1,1) Darcy
apr = 100*barsa*G.cells.centroids(:,1)*1/nh;
clf
%% Plotting the pressure solutions

 figure
 subplot (3,2,1)
 plotCellData(G, convertTo(resSol2.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 title('Mimetic')
 xlabel('x'), ylabel('y'), zlabel('Depth');
 view(3); shading faceted; camproj perspective; axis tight;
 cax = caxis; colorbar
 
 subplot (3,2,2)
 plotCellData(G, convertTo(resSol3.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 title('Hybrid mimetic')
 xlabel('x'), ylabel('y'), zlabel('Depth');
 view(3); shading faceted; camproj perspective; axis tight;
 caxis(cax); colorbar
 
 subplot (3,2,3)
 plotCellData(G, convertTo(resSol4.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 title('MPFA')
 xlabel('x'), ylabel('y'), zlabel('Depth');
 view(3); shading faceted; camproj perspective; axis tight;
 caxis(cax); colorbar
 
subplot(3,2,4)
 plotCellData(G, convertTo(resSol5.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 title('Hybrid MPFA')
 xlabel('x'), ylabel('y'), zlabel('Depth');
 view(3); shading faceted; camproj perspective; axis tight;
 caxis(cax); colorbar

  subplot (3,2,5)
 plotCellData(G, convertTo(resSol1.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 title('TPFA')
 xlabel('x'), ylabel('y'), zlabel('Depth');
 view(3); shading faceted; camproj perspective; axis tight;
 caxis(cax); colorbar
 
subplot(3,2,6)
 plotCellData(G, convertTo(apr, barsa()), ...
              'EdgeColor', 'k');
 title('Analytical')
 xlabel('x'), ylabel('y'), zlabel('Depth');
 view(3); shading faceted; camproj perspective; axis tight;
 caxis(cax); colorbar

 
%%
% Calculating the L^2 error. 

errtwot = sqrt(sum((convertTo((resSol1.pressure -apr),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errExact = sqrt(sum((convertTo(apr,barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwom = sqrt(sum((convertTo((resSol2.pressure - apr),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwoh = sqrt(sum((convertTo((resSol3.pressure - apr),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwomp = sqrt(sum((convertTo((resSol4.pressure - apr),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwohmp = sqrt(sum((convertTo((resSol5.pressure - apr),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));   
   
% Printing the relative L^2 error
fprintf(1, 'TPFA:\t%f\n', errtwot/errExact);
fprintf(1, 'MIMETIC:\t%f\n', errtwom/errExact);
fprintf(1, 'HYBRID MIMETIC:\t%f\n', errtwoh/errExact); 
fprintf(1, 'MPFA:\t%f\n', errtwomp/errExact); 
fprintf(1, 'HYBRID MPFA:\t%f\n', errtwohmp/errExact); 

displayEndOfDemoMessage(mfilename)
