% A 2.5D pebi grid is made and this is extruded to three 
% dimensions. This example is made to compare the solutions obtained by
% TPFA, MPFA, Mimetic using inner product 'ip_simple', the hybrid mimetic
% method and the hybrid MPFA method to the analytic solution.

% Setting the dimension of the grid
N = 7; M = 5; zval = 5;

% Generating the points that make up the grid.
[x,y] = ndgrid(0:N,0:M);

% Changing the points such that the grid is not completely regular.
x(2:N,2:M) = x(2:N,2:M) + 0.3*randn(N-1,M-1);
y(2:N,2:M) = y(2:N,2:M) + 0.3*randn(N-1,M-1);

% Triangulate the grid and then use the function pebi to make a 2.5D pebi
% grid.
aG = pebi(triangleGrid([x(:) y(:)]));

% Extrude the grid to three dimensions.
G = makeLayeredGrid(aG,zval);

% We continue to compute centroids and volumes of the cells and centroids, 
% normals, and areas for the faces. 
G = computeGeometry(G);

% Plot the grid
plotGrid(G)

%% Set rock and fluid data

% We set the permeability, rock.perm, to be homogeneous and isotropic 
% K = diag(1,1,1) Darcy. 
rock.perm = repmat([1000, 1000, 1000].* milli*darcy(), [G.cells.num, 1]);

% The viscosity is specified by saying that the reservoir is filled with a 
% single fluid, where the viscosity value is 1 cp. The density of the fluid 
% is set to be 1000 kg/m^3.
fluid = initSingleFluid('mu', 1*centi*poise, 'rho', 1000*kilogram/meter^3);


%% Initialize reservoir simulator

% We set the initial reservoir pressure to be 0. The single-phase
% saturation is also set to 0, since this is the default behaviour of
% initResSol.
xi = initResSol(G, 0);

%% Impose Dirichlet boundary conditions
westFaces = find(G.faces.centroids(:,1) == 0);
bc = addBC([], westFaces, 'pressure', 0);
maxx = max(G.faces.centroids(:,1));
eastFaces = find(G.faces.centroids(:,1) >= N);
bc = addBC(bc, eastFaces, 'pressure', 100*barsa); 

% Calculating the analytic pressure solution. Valid in current form when 
% rock.perm = diag(1,1,1) Darcy
pa = 100*barsa*G.cells.centroids(:,1)*1/N;
gravity off;

%% Construct linear system

% Hybrid mimetic
H = computeHybridMimeticIP(G, rock, 'InnerProduct', 'ip_hyb');
% TPFA
T = computeTrans(G, rock);
% Mimetic
M = computeMimeticIP(G, rock, 'InnerProduct', 'ip_simple');

% Adding the module where the MPFA function is.
mrstModule add mpfa


% MPFA
MP = computeMultiPointTrans(G, rock);
% Hybrid MPFA
HMP = computeHybridMultiPointTrans(G, rock);

%% Solve the linear system

% Solve Hybrid mimetic system.
xh = solveIncompFlow(xi, G, H, fluid, 'bc', bc);
% Solve TPFA system
xt = incompTPFA(xi, G, T, fluid, 'bc', bc);
% Solve Mimetic system
xm = solveIncompFlow(xi, G, M, fluid, 'bc', bc);
% Solve MPFA system
xmp = incompMPFA(xi, G, MP, fluid, 'bc', bc);
% Solve hybrid MPFA system
xhmp = incompMPFA(xi, G, HMP, fluid, 'bc', bc);


%% Plot solution    

% Plotting the pressure solutions
 figure,
 subplot (3,2,1), 
 plotCellData(G, convertTo(xm.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 view(3); shading faceted; camproj perspective; axis tight off;
 cax = caxis; colorbar, title('Mimetic');

 
 subplot (3,2,2),  
 plotCellData(G, convertTo(xh.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 view(3); shading faceted; camproj perspective; axis tight off;
 caxis(cax); colorbar, title('Hybrid mimetic');

 subplot (3,2,3), 
 plotCellData(G, convertTo(xmp.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 view(3); shading faceted; camproj perspective; axis tight off;
 caxis(cax); colorbar, title('MPFA');

 
 subplot (3,2,4),  
 plotCellData(G, convertTo(xhmp.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 view(3); shading faceted; camproj perspective; axis tight off;
 caxis(cax); colorbar, title('Hybrid MPFA');

 
 subplot (3,2,5),  
 title('hybrid');
 plotCellData(G, convertTo(xt.pressure(1:G.cells.num), barsa()), ...
              'EdgeColor', 'k');
 view(3); shading faceted; camproj perspective; axis tight off;
 caxis(cax); colorbar, title('TPFA')

 
subplot(3,2,6),  
plotCellData(G, convertTo(pa, barsa()), ...
              'EdgeColor', 'k');
 view(3); shading faceted; camproj perspective; axis tight off;
 caxis(cax); colorbar, title('Analytical');  
 
    
%% Calculate the  L2 error

errtwot = sqrt(sum((convertTo((xt.pressure -pa),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errExact = sqrt(sum((convertTo(pa,barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwom = sqrt(sum((convertTo((xm.pressure - pa),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwoh = sqrt(sum((convertTo((xh.pressure - pa),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwomp = sqrt(sum((convertTo((xmp.pressure - pa),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));
errtwohmp = sqrt(sum((convertTo((xhmp.pressure - pa),barsa)).^2.*G.cells.volumes)./...
       sum(G.cells.volumes));   
   
% Print the relative L^2 error.
fprintf(1, 'IP_TPF:\t%f\n', errtwot/errExact);
fprintf(1, 'MIMETIC:\t%f\n', errtwom/errExact);
fprintf(1, 'HYBRID MIMETIC:\t%f\n', errtwoh/errExact); 
fprintf(1, 'MPFA:\t%f\n', errtwomp/errExact); 
fprintf(1, 'HYBRID MPFA:\t%f\n', errtwohmp/errExact); 
   
   
   
   
   
