% Example to show how the hybrid methods perform when we have a grid with
% skew pillars, but the ratio DX/DZ is large. DX is the length in the x-direction,
% adn DZ is the length in the z-direction. We start with a relative small
% ratio, and then increase it to find out if it is sufficient to  use a
% hybrid method applying TPFA in the vertical direction.

% ratio = DX/DZ
ratio = [2 5 10 50 100];

% Initializing the matrices where we want to place the flux values.
flux1 = zeros(5,1);
flux2 = zeros(5,1);
flux3 = zeros(5,1);
flux4 = zeros(5,1);
flux5 = zeros(5,1);
fluxan = zeros(5,1);

% Looping over the different ratios.
for i = 1:length(ratio)
%% Making the grid    

% Setting physical dimensions 25*DX/DZ x 1 x 25    
physDims   = [25*ratio(i), 1, 25];
% Setting the dimensions of the grid, i.e., we have a grid that is divided
% into 5x1x5 grid blocks.
dims = [5 1 5];

% Making the grid
g.cartDims = reshape(dims, 1, []);
[X, Y, Z]  = ndgrid(linspace(0, physDims(1), dims(1) + 1), ...
                    linspace(0, physDims(2), dims(2) + 1), ...
                    linspace(0, physDims(3), dims(3) + 1));

% Make skew pillars and perturb the layers.
%X = X + 0.2*(0.5 - abs(X - 0.5)).*(Z - 0.5);
X(2,:,:) = X(2,:,:) - Z(2,:,:)/tand(80);
X(3,:,:) = X(3,:,:) - Z(3,:,:)/tand(70);
X(4,:,:) = X(4,:,:) - Z(4,:,:)/tand(55);
X(5,:,:) = X(5,:,:) - Z(5,:,:)/tand(45);

Z = sort(Z, 3);


% Make pillars
n = prod(dims(1:2) + 1);
lines = zeros([n, 6]);
lines(:, [1, 4]) = reshape(X(:,:,[1, end]), [n, 2]);
lines(:, [2, 5]) = reshape(Y(:,:,[1, end]), [n, 2]);
lines(:, [3, 6]) = reshape(Z(:,:,[1, end]), [n, 2]);
g.COORD = reshape(lines.', [], 1);

% Assign z-coordinates
ind = @(d) 1 + fix((1 : 2*dims(d)) ./ 2);
z   = Z(ind(1), ind(2), ind(3));

g.ZCORN = z(:);

% Assign active cells
g.ACTNUM = reshape(ones(dims, 'int32'), [], 1);

% Process the grid which is made as a grdecl. 
G = processGRDECL(g);

% We continue to compute centroids and volumes of the cells and centroids, 
% normals, and areas for the faces. 
G = computeGeometry(G);


%% Setting fluid and rock properties

% We set the permeability, rock.perm, to be homogeneous and isotropic 
% K = diag(1,1,1) Darcy. 
rock.perm = repmat([1000, 1000, 1000].* milli*darcy(), [G.cells.num, 1]);

% The viscosity is specified by saying that the reservoir is filled with a 
% single fluid, where the viscosity value is 1 cp. The density of the fluid 
% is set to be 1014 kg/m^3.
fluid     = initSingleFluid('mu' ,    1*centi*poise     , ...
                           'rho', 1014*kilogram/meter^3);
                        
%% Impose Dirichlet boundary conditions
westFaces = find(G.faces.centroids(:,1) == 0);
bc = addBC([], westFaces, 'pressure', 0);
xmax = max(G.faces.centroids(:,1));
eastFaces = find(G.faces.centroids(:,1) == xmax);
bc = addBC(bc, eastFaces, 'pressure', 3000*barsa);   

% Calculating the analytic solution of the flux
aflux =-(darcy*3000*barsa/(centi*poise))*G.faces.normals(:,1)/physDims(1);

%% Initialize reservoir simulator

% We set the initial reservoir pressure to be 0. The single-phase
% saturation is also set to 0, since this is the default behaviour of
% initResSol.
resSol = initResSol(G, 0.0);

%% Construct and solve linear system

% Hybrid MPFA
rh = computeHybridMimeticIP(G, rock, 'InnerProduct', 'ip_hyb');
resSol1 = solveIncompFlow(resSol, G, rh, fluid, ...
                         'bc', bc, 'MatrixOutput', true);

% Mimetic                     
rm = computeMimeticIP(G, rock, 'InnerProduct', 'ip_simple');
resSol2 = solveIncompFlow(resSol, G, rm, fluid, ...
                         'bc', bc, 'MatrixOutput', true);

% TPFA                     
T=computeTrans(G,rock,'K_system','loc_xyz');
resSol3 = incompTPFA(resSol,G,T,fluid,'bc',bc);

% Adding the module where the MPFA function is.
mrstModule add mpfa

% MPFA
rmp = computeMultiPointTrans(G, rock);
resSol4 = incompMPFA(resSol, G, rmp, fluid, 'bc', bc);

% Hybrid MPFA
rhmp = computeHybridMultiPointTrans(G, rock);
resSol5 = incompMPFA(resSol, G, rhmp, fluid, 'bc', bc);

% Calculating the total flux obtained by the different methods and plaing
% them in matrices. These are the values we want to compare.

flux1(i) = sum(abs(resSol1.flux(eastFaces)));
flux2(i) = sum(abs(resSol2.flux(eastFaces)));
flux3(i) = sum(abs(resSol3.flux(eastFaces)));
flux4(i) = sum(abs(resSol4.flux(eastFaces)));
flux5(i) = sum(abs(resSol5.flux(eastFaces)));
fluxan(i) = sum(abs(aflux(eastFaces)));

end


 