% We make a grid with physical dimension 1000x500x20 where we have 10x5x5
% grid blocks. The grid is twisted in the xy-plane. We want to find out 
% when non-physical cycles occur for MPFA and the mimetic method. We have 
% an anisotropic permeability ratio r = permX/permZ. We change this ratio 
% according to the numbers in the vector fak = permZ/permX, to see if we 
% get cycles when the ratio changes. We also rotate the grid and for each 
% new angle we see what happens when the anisotropic permesbility rate is 
% increased. 


% fak is the number we want permZ/permX to be.
fak = [1 1/2 1/5 1/10 1/100 1/1000 1/1000000000000000];
% The angle of the slope of the grid
deg = [0 1 2 5 10];

% Initializing the matrices where we want to state if the solution has
% cycles or not
cyc1 = zeros(length(fak), length(deg));
cyc2 = zeros(length(fak), length(deg));
cyc3 = zeros(length(fak), length(deg));
cyc4 = zeros(length(fak), length(deg));
cyc5 = zeros(length(fak), length(deg));

% Initializing the matrices where we want to place the total value of the
% flux for each case.
flmat1 = zeros(length(fak), length(deg));
flmat2 = zeros(length(fak), length(deg));
flmat3 = zeros(length(fak), length(deg));
flmat4 = zeros(length(fak), length(deg));
flmat5 = zeros(length(fak), length(deg));

% Initializing the matrices where we want to place the ratio of cells in
% cycles = R, the number of cycles = N and the size of the largest cycle =
% nmax. 
R1mat= zeros(length(fak), length(deg));
N1mat = zeros(length(fak), length(deg));
nmax1mat = zeros(length(fak), length(deg));

R2mat = zeros(length(fak), length(deg));
N2mat = zeros(length(fak), length(deg));
nmax2mat = zeros(length(fak), length(deg));

R3mat = zeros(length(fak), length(deg));
N3mat = zeros(length(fak), length(deg));
nmax3mat = zeros(length(fak), length(deg));

R4mat = zeros(length(fak), length(deg));
N4mat = zeros(length(fak), length(deg));
nmax4mat = zeros(length(fak), length(deg));

R5mat = zeros(length(fak), length(deg));
N5mat = zeros(length(fak), length(deg));
nmax5mat = zeros(length(fak), length(deg));

% Looping over the different cases where the rotation angle and the
% permeability rate is changed.
for i = 1:length(deg)
    
    % Setting the dimensions of the grid and
    G = cartGrid([10 5 5], [1000, 500, 20]);
    
    % We continue to compute centroids and volumes of the cells and centroids, 
    % normals, and areas for the faces. 
    G = computeGeometry(G);
    
    % The grid is perturbed in the xy-plane.
    G.nodes.coords(:,2) = G.nodes.coords(:,2) + 0.08*G.nodes.coords(:,1);
    
    % The z-value is changed according to the degrees, i.e., the grid is 
    % rotaed.
    G.nodes.coords(:,3) = G.nodes.coords(:,3) + tand(deg(i))*G.nodes.coords(:,1);
    
    % We need to compute the geometry again since we have changed the grid.
    G=computeGeometry(G);

    for j = 1:length(fak)       
        %% Setting fluid and rock properties 
        
        % We set the permeability, rock.perm, to be homogeneous and 
        % anisotropic K = diag(1,1,1*fak) Darcy. 
        rock.perm = repmat([1000, 1000, 1000*fak(j)].* milli*darcy(), [G.cells.num, 1]);

        % The viscosity is specified by saying that the reservoir is filled 
        % with a single fluid, where the viscosity value is 1 cp. The 
        % density of the fluid is set to be 1014 kg/m^3.
        fluid     = initSingleFluid('mu' ,    1*centi*poise     , ...
                            'rho', 1014*kilogram/meter^3);
        
        % Setting boundary conditions                
        westFaces = find(G.faces.centroids(:,1) == 0);
        bc = addBC([], westFaces, 'pressure', 0);
        xmax = max(G.faces.centroids(:,1));
        eastFaces = find(G.faces.centroids(:,1) == xmax);
        bc = addBC(bc, eastFaces, 'pressure', 100*barsa);   

        gravity off;
        
        %% Initialize reservoir simulator

        % We set the initial reservoir pressure to be 0. The single-phase
        % saturation is also set to 0, since this is the default behaviour 
        % of initResSol.
        resSol = initResSol(G, 0.0);
        
        %% Construct and solve linear system
        
        % Hybrid mimetic
        rh = computeHybridMimeticIP(G, rock, 'InnerProduct', 'ip_hyb');
        resSol1 = solveIncompFlow(resSol, G, rh, fluid, ...
                          'bc', bc, 'MatrixOutput', true);

        % Adding the module where the MPFA function is.
        mrstModule add mpfa

        % Hybrid MPFA
        rh = computeHybridMultiPointTrans(G,rock);
        resSol2 = incompMPFA(resSol, G, rh, fluid, 'bc', bc, 'MatrixOutput', true);
        
        % MPFA
        rm = computeMultiPointTrans(G,rock);
        resSol3 = incompMPFA(resSol, G, rm, fluid, 'bc', bc, 'MatrixOutput', true);
        
        % Mimetic
        rm = computeMimeticIP(G, rock, 'InnerProduct', 'ip_simple');
        resSol4 = solveIncompFlow(resSol, G, rm, fluid, ...
                          'bc', bc, 'MatrixOutput', true);
        
        %TPFA
        T=computeTrans(G,rock,'K_system','loc_xyz');
        resSol5 = incompTPFA(resSol,G,T,fluid,'bc',bc, 'MatrixOutput', true);


        % Placing the total flux-values here
        flmat1(j,i) = sum(abs(resSol1.flux(eastFaces)));
        flmat2(j,i) = sum(abs(resSol2.flux(eastFaces)));
        flmat3(j,i) = sum(abs(resSol3.flux(eastFaces)));
        flmat4(j,i) = sum(abs(resSol3.flux(eastFaces)));
        flmat5(j,i) = sum(abs(resSol3.flux(eastFaces)));
        
        % Finding the ratio of cells in cycles, the number of cycles and
        % the maximum cycle.
        [R1, N1, nmax1, cyclefree1] = findingCycles(G, resSol1.flux);
        [R2, N2, nmax2, cyclefree2] = findingCycles(G, resSol2.flux);
        [R3, N3, nmax3, cyclefree3] = findingCycles(G, resSol3.flux);
        [R4, N4, nmax4, cyclefree4] = findingCycles(G, resSol4.flux);
        [R5, N5, nmax5, cyclefree5] = findingCycles(G, resSol5.flux);

        % Stating whether the solution contains cycles or not. 
        if(cyclefree1 == true)
            cyc1(j,i) = 1;
        end
        
        if(cyclefree2 == true)
            cyc2(j,i) = 1;
        end
        
        if(cyclefree3 == true)
            cyc3(j,i) = 1;
        end
        
        if(cyclefree4 == true)
            cyc4(j,i) = 1;
        end
        
        if(cyclefree5 == true)
            cyc5(j,i) = 1;
        end
        
        % Placing the ratio of cells in cycles, the number of cyles and the
        % size of the largest cycle in their respective matrices.
        R1mat(j,i)= R1;
        N1mat(j,i) = N1;
        nmax1mat(j,i) = nmax1;
        
        R2mat(j,i)= R2;
        N2mat(j,i) = N2;
        nmax2mat(j,i) = nmax2;
        
        R3mat(j,i)= R3;
        N3mat(j,i) = N3;
        nmax3mat(j,i) = nmax3;
        
        R4mat(j,i)= R4;
        N4mat(j,i) = N4;
        nmax4mat(j,i) = nmax4;
        
        R5mat(j,i)= R5;
        N5mat(j,i) = N5;
        nmax5mat(j,i) = nmax5;


    end
end 