%% Big pebi grid
% This grid is found in the module scriptsMRSTpaper under example6, but it 
% is also added to this module. The grid is a 2.5D pebi grid containing a 
% curved fault: We use the hybrid mimetic, hybrid mpfa, mpfa, and mimetic 
% methods and consider non-physical cycles occuring when we have 
% anisotropic permeabilities and these methods are used. We wantg to find
% out the ratio of cells in cycles, the number of cycles and the size of
% the largest cycle. 


%% Set up parameters
maxArea     = 4;                % controlling resolution of fine grid
[az,el]     = deal(-42,68);     % view point for plots
useTriangle = false;            % should we use Triangle?
useMETIS    = false;            % should we use METIS?
metisPath   = fullfile('myLocalMETISpath'); % change to local METIS path
vargs       = {'EdgeColor', 'k', 'EdgeAlpha', .05};

%% Read grid from file
 dd = fileparts(mfilename('fullpath'));
 load(fullfile(dd, 'data', 'pebi'));
 clear dd;
 G = Gp; clear Gp;
 
 % We continue to compute centroids and volumes of the cells and centroids, 
 % normals, and areas for the faces. 
 G = computeGeometry(G);

 % Plot the grid
   figure
   plotGrid(G)
   
   % We set the permeability, rock.perm, to be homogeneous and anisotropic,
   % K = diag(1,1,0.01) Darcy. The porosity is 0.3. 
   rock = struct('perm', repmat([1000, 1000, 10].* milli*darcy(), [G.cells.num, 1]), ...
                 'poro', repmat(0.3,             [G.cells.num, 1]));
             
   % The viscosity is specified by saying that the reservoir is filled 
   % with a single fluid, where the viscosity value is 1 cp. The 
   % density of the fluid is set to be 1000 kg/m^3.          
   fluid = initSingleFluid('mu', 1*centi*poise, 'rho', 1000*kilogram/meter^3);          
   gravity off;


   %% Assign boundary conditions
   % West sides (high pressure)
   d  = abs(G.faces.centroids(:,1) - min(G.faces.centroids(:,1)));
   ws = find (d < 1e6*eps);

   % East sides (low pressure)
   d  = abs(G.faces.centroids(:,1) - max(G.faces.centroids(:,1)));
   es = find (d < 1e6*eps);

   bc = addBC([], ws, 'pressure', 250*barsa);
   bc = addBC(bc, es, 'pressure', 100*barsa);

   %% Discretize and solve single-phase system
   
   % Initialize pressure system
   xi = initResSol(G, 0);
   
   % Hybrid mimetic
   H     = computeHybridMimeticIP(G, rock, 'InnerProduct', 'ip_hyb');
   xh = solveIncompFlow(xi, G, H, fluid, 'bc', bc);
   
   % Mimetic
   M = computeMimeticIP(G, rock, 'InnerProduct', 'ip_simple');
   xm = solveIncompFlow(xi, G, M, fluid, 'bc', bc);
  
   % Adding the module where the MPFA function is.
   mrstModule add mpfa
   
   % MPFA
   MPFA = computeMultiPointTrans(G, rock);
   xmpfa = incompMPFA(xi, G, MPFA, fluid, 'bc', bc);
   
   % Hybrid MPFA
   Mpfa = computeHybridMultiPointTrans(G, rock);
   xhmpfa = incompMPFA(xi, G, Mpfa, fluid, 'bc', bc);
%% Plot pressure solution 
   
   figure
   subplot(2,2,1)
   plotCellData(G, convertTo(xm.pressure, barsa), vargs{:});
   axis tight, view(3), colorbar
   title('Mimetic')
   
   subplot(2,2,2)
   plotCellData(G, convertTo(xh.pressure, barsa), vargs{:});
   axis tight, view(3), colorbar
   title('Hybrid mimetic')
   
   subplot(2,2,3)
   plotCellData(G, convertTo(xmpfa.pressure, barsa), vargs{:});
   axis tight, view(3), colorbar
   title('MPFA')
   
   subplot(2,2,4)
   plotCellData(G, convertTo(xhmpfa.pressure, barsa), vargs{:});
   axis tight, view(3), colorbar
   title('Hybrid MPFA')
%% Finding the ratio of cells in cycles, the number of cycles and the maximum cycle.

% R is the ratio of cells in cycles. N is the number of cycles, nmax is the
% size of the largest cycle. cyclefree is true if the solution contains no
% cycles and false otherwise.
[R1, N1, nmax1, cyclefree1] = findingCycles(G, xh.flux);
[R2, N2, nmax2, cyclefree2] = findingCycles(G, xm.flux);
[R3, N3, nmax3, cyclefree3] = findingCycles(G, xmpfa.flux);
[R4, N4, nmax4, cyclefree4] = findingCycles(G, xhmpfa.flux);

% Plotting the flux-matrix and the cycles. 
u1 = visualizeCycles(G, xm.flux);
u2 = visualizeCycles(G, xmpfa.flux);

