function T = computeHybridMultiPointTrans(g, rock, varargin)
%Compute hybrid multi-point transmissibilities.
%
% SYNOPSIS:
%   T = computeMultiPointTrans(G, rock)
%
% REQUIRED PARAMETERS:
%   G       - Grid data structure as described by grid_structure.
%
%   rock    - Rock data structure with valid field 'perm'.  The
%             permeability is assumed to be in measured in units of
%             metres squared (m^2).  Use function 'darcy' to convert from
%             (milli)darcies to m^2, e.g.,
%
%                 perm = convertFrom(perm, milli*darcy)
%
%             if the permeability is provided in units of millidarcies.
%
%             The field rock.perm may have ONE column for a scalar
%             permeability in each cell, TWO/THREE columns for a diagonal
%             permeability in each cell (in 2/3 D) and THREE/SIX columns
%             for a symmetric full tensor permeability.  In the latter
%             case, each cell gets the permeability tensor
%
%                 K_i = [ k1  k2 ]      in two space dimensions
%                       [ k2  k3 ]
%
%                 K_i = [ k1  k2  k3 ]  in three space dimensions
%                       [ k2  k4  k5 ]
%                       [ k3  k5  k6 ]
%
% OPTIONAL PARAMETERS
%   verbose   - Whether or not to emit informational messages throughout the
%               computational process.  Default value depending on the
%               settings of function 'mrstVerbose'.
%
%   facetrans - 
%
% RETURNS:
%   T - half-transmissibilities for each local face of each grid cell
%       in the grid.  The number of half-transmissibilities equal the
%       number of rows in G.cells.faces.
%
% COMMENTS:
%   PLEASE NOTE: Face normals have length equal to face areas.
%
% SEE ALSO:
%   incompMPFA, mrstVerbose.


%{
Copyright 2012 Christine M. Ø. Haugland.

This file is part of the Hybrid Mimetic and Multipoint module(HybridMethods).
HybridMethods is a module for the MATLAB Reservoir Simulation Toolbox (MRST).

HybridMethods is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HybridMethods is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HybridMethods.  If not, see <http://www.gnu.org/licenses/>.

%} 

% This file is a small modification of the file computeMultiPointTrans.m. 
% The only thing that is added by the above author is the function 
% createNewMapping and the lines between  
% B = computeLocalFluxMimeticIP(g, rock, cno, fno, nno, subhfno); and 
% B = processFaceTrans(B, g, opt.facetrans(:,1), opt.facetrans(:,2), fno);. 
% The function computeMultiPointTrans.m is under the following copyright. 

%{
Copyright 2009, 2010, 2011 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Date: 2011-12-31 17:43:06 +0100 (Sat, 31 Dec 2011) $
% $Revision: 8727 $

% Written by Jostein R. Natvig, SINTEF ICT, 2009.

   opt = struct('verbose',   mrstVerbose,   ...
                'facetrans', zeros([0, 2]));
   opt = merge_options(opt, varargin{:});

   if opt.verbose, 
      fprintf('Computing mappings between for subfaces ...\t');
      t0 = tic;
   else
      t0 = [];
   end

   %% 
   % Enumerate sub-faces and sub-half-faces.  
   % Create mappings from {cells, nodes, faces} to {subfaces, subhalffaces}.
   
   [cno, nno, hfno, fno, subfno, subhfno] = createMapping(g);
   
   % This function is new compared to computeMultoPointTrans.m. the
   % function marks which faces that are top or bottom. If a face i is top or
   % bottom then skille(i) = 1.
   [cno, nno, hfno, fno, subfno, subhfno, skille] = ...
       createNewMapping(g, cno, nno, hfno, fno, subfno, subhfno);

   tocif(opt.verbose, t0);

   if opt.verbose,
      fprintf('Computing inner product on sub-half-faces ...\t');
      t0 = tic;
   end
   
   B = computeLocalFluxMimeticIP(g, rock, cno, fno, nno, subhfno);

   
   % Compute TPFA transmissibilities for the top and bottom faces.
   halfsubchange = subhfno(logical(skille));
   hfnotosubhfno = [hfno(logical(skille)) subhfno(logical(skille))];
   TPFA = computeTrans(g, rock);
   
   numnodes = double(diff(g.faces.nodePos));
   halffacetoface = [fno(logical(skille)) subhfno(logical(skille))];
   i = (1:length(halffacetoface))';
   facenodes(i) = numnodes(halffacetoface(i,1));
   
   % Setting the entries computed by MPFA belonging to top and bottom faces to 0.
   B(:,halfsubchange) = 0;
   B(halfsubchange,:) = 0;

   % Replacing the entries with TPFA transmissibilities.
   for i = 1:length(halfsubchange)
       B(halfsubchange(i),halfsubchange(i)) = 1/(TPFA(hfnotosubhfno(i,1))/facenodes(i));
   end
   
   B = processFaceTrans(B, g, opt.facetrans(:,1), opt.facetrans(:,2), fno);
   
   tocif(opt.verbose, t0);

   if opt.verbose,
      fprintf('Computing inverse mixed innerproduct ...\t');
      t0 = tic;
   end
   
   %% Create matrices needed to compute transmissibilities
   s     = 2*(cno==g.faces.neighbors(fno,1))-1;
   D     = sparse(subhfno, subfno, 1); % Hybrid D matrix
   Do    = sparse(subhfno, subfno, s); % Mixed  D matrix
   C     = sparse(subhfno, cno, 1);
   
   % c1 adds up sub-half-face contributions for each half-face
   c1     = sparse(subhfno, hfno, 1);
   
   % d1 adds up sub-face contributions for each face.
   d1     = sparse(subfno, fno, 1);
   
   % Note that c1'*D*d1 is equal to the reglar mimetic D matrix.
   
   %% Construct the inverse of the mixed B
   % In the local-flux mimetic method, the mixed mass matrix DoBDo is block
   % diagonal with n_i x n_i blocks due to the special form of the hybrid
   % mass matrix.  Here n_i is the number of cells adjacent to a node in
   % the grid.

   DoBDo = Do'*B*Do;

   %% Invert DoBDo   
   tmp      = unique([nno, subfno], 'rows');
   p        = tmp(:,2);
   P        = sparse(1:numel(p), p, 1, numel(p), numel(p));
   [sz, sz] = rlencode(tmp(:,1));
   iDoBDo   = P'*invertDiagonalBlocks(DoBDo(p, p), sz)*P;
   clear tmp sz P p 
   
   tocif(opt.verbose, t0);
   
   if opt.verbose,
      fprintf('Computing multi-point transmissibilities ...\t');
      t0 = tic;
   end
      
   
   %% Compute multi-point transmissibilities
   % for all faces in terms of cell pressures and outer boundary pressures.
   b = full(sum(D, 1)) == 1;
   %T = c1'*Dm*inv(Dm'*B*Dm)*Dm'*[C, -D(:,b)*d1(b,:)];
    %if(nargout==2)
   Tg=c1'*Do*iDoBDo*Do';
    %end
    

   T = Tg*[C, -D(:,b)*d1(b,:)];
   Tg=Tg*c1;
   tocif(opt.verbose, t0);
   % c1'*D*d1 har samme struktur som vanlig D.
   % T er feil størrelse å returnere dersom gravitasjon skal håndteres
   % skikkelig.  Gravitasjonsleddet kommer inn som c1'*Dm*iDmBDm*Dm'*f.
   % Siden gravitasjonsbidraget for all subfaces er likt kan de sikkert
   % skrives om til c1'*Dm*iDmBDm*Dm'*F*g der F*G=f.
   T=struct('T',T,'Tg',Tg);
   %T(:,all(T==0, 1))=[];
end

function [cno, nno, hfno, fno, subfno, subhfno] = createMapping(g)
   %% Create mapping from sub-half-face to cell, node, face, half-face and
   %% sub-face
   cellno   = rldecode(1:g.cells.num, diff(g.cells.facePos), 2) .';
   col      = 1+(cellno == g.faces.neighbors(g.cells.faces(:,1), 2));
   nhfaces  = g.cells.facePos(end)-1;
   hfaces   = accumarray([g.cells.faces(:,1), col], 1:nhfaces);
   hfaces   = rldecode(hfaces, diff(g.faces.nodePos));

   
   cells    =  rldecode(g.faces.neighbors, diff(g.faces.nodePos));
   nodes    =  repmat(g.faces.nodes, [2,1]);
   faces    =  repmat(rldecode(1:g.faces.num, diff(g.faces.nodePos),2)', [2,1]);
   subfaces =  repmat((1:size(g.faces.nodes,1))', [2,1]);
   i        =  cells~=0;
   w        =  [cells(i), nodes(i), hfaces(i), faces(i), subfaces(i)];
   w        =  double(sortrows(w));
   
   
   cno     = w(:,1); 
   nno     = w(:,2);
   hfno    = w(:,3);
   fno     = w(:,4);
   subfno  = w(:,5);
   subhfno = (1:numel(cno))';
end


function [cno, nno, hfno, fno, subfno, subhfno, skille] = createNewMapping(g, cno, nno, hfno, fno, subfno, subhfno)
   w(:,1) = cno;
   w(:,2) = nno;
   w(:,3) = hfno;
   w(:,4) = fno;
   w(:,5) = subfno;
   
   w(:,6) = zeros(length(w),1);
   w(:,7) = zeros(length(w),1);
   
   
   
   cf = double(g.cells.faces(:,1));

   cellno = rldecode(1 : g.cells.num, diff(g.cells.facePos), 2) .';
   sgn    = 2*double(g.faces.neighbors(cf, 1) == cellno) - 1;

   areas      = g.faces.areas    (cf  );
   centVecs   = g.faces.centroids(cf,:) - g.cells.centroids(cellno,:);
   outNormals = bsxfun(@times, g.faces.normals(cf,:), sgn);
   
   dimProd = double(diff(g.cells.facePos));
  

   ix = 0;
   for i = 1:g.cells.num
       nF  = dimProd(i);
       pR  = ix  + 1 : ix  + nF;
       
       a = areas(pR);
       N = outNormals(pR,:);
       
       
       fakeN = bsxfun(@rdivide, N, a);
   
       
        cellfaces = zeros(g.cells.facePos(i+1)-g.cells.facePos(i),2);
% 
        for j = (g.cells.facePos(i):g.cells.facePos(i+1)-1)
           cellfaces(j-g.cells.facePos(i)+1,1) = g.cells.faces(j); 
        end
        cellfaces(:,2) = fakeN(:,3);
        maxc = max(cellfaces(:,2));
        minc = min(cellfaces(:,2));
% 
        ta = cellfaces(:,2) == maxc;
        tb = cellfaces(:,2) == minc;
        top = cellfaces(ta,1);
        bottom = cellfaces(tb,1);

       wcells = w(:,1) == i;
       w(wcells,6) = w(wcells,4) == top;
       w(wcells,7) = w(wcells,4) == bottom; 
       w(wcells, 8) = w(wcells,6)+w(wcells,7);
       ix = ix + nF;
   end 
   
   cno     = w(:,1); 
   nno     = w(:,2);
   hfno    = w(:,3);
   fno     = w(:,4);
   subfno  = w(:,5);
   subhfno = (1:numel(cno))';
   skille = w(:,8);
   
end


   
function B = computeLocalFluxMimeticIP(g, rock, cno, fno, nno, subhfno)
   [a, blocksz] = rlencode([cno,nno]);
   dims         = size(g.nodes.coords, 2);
   assert(all(blocksz==dims));
   %% Expand centroid differences, normals, signs and permeabilities to
   %  block-diagonal matrices such that B may be constructed by matrix-matrix
   %  multiplications:
   [i,j] = ndgrid(subhfno, 1:dims);
   j     = bsxfun(@plus, j, rldecode(cumsum(blocksz)-blocksz(1), blocksz));
   
   % Use original face centroids and cell centroids, NOT actual subface
   % centroids.  This corresponds to an MPFA method (O-method)
   R     = g.faces.centroids(fno,:) - g.cells.centroids(cno,:);
   R     = sparse(i,j,R);

   % Subface sign == face sign
   sgn   = 2*(cno == g.faces.neighbors(fno,1)) -1;
   
   % Use fraction of face normal as subface normal
   numnodes = double(diff(g.faces.nodePos));
   N     = g.faces.normals  (fno,:).*sgn(:,ones(1, dims))./ ...
           numnodes(fno, ones(1,dims));
   N     = sparse(i,j,N);
   
   k     = permTensor(rock, dims);

   assert (size(k,1) == g.cells.num, ...
          ['Permeability must be defined in active cells only.\n', ...
           'Got %d tensors, expected %d (== number of cells).'],   ...
           size(k,1), g.cells.num);

   K     = sparse(i,j,reshape(k(a(:,1),:)', dims, [])');
   clear k a blocksz sgn
   
   %% Construct B : Invert diagonal blocks of size sz
   d     = size(g.nodes.coords,2);       % == dims
   sz    = repmat(d, [size(R, 1)/d, 1]);
   B     = R*invertDiagonalBlocks(N*K, sz);
end

function B = processFaceTrans(B, g, f, t, fno)
%% Modify B to account for face transmissibility (t) for regular faces (f)
% (a) Distribute 2*t to each half-face such that the harmonic mean
%     equals t.  (Use 1*t for boundary faces)
%
% (b) As an approximation, divide facetrans by number of corners in face
%     to distribute face transmissibility on each sub-face.
%
% (c) Modify B <-- B + diag(1/(2*t)).  This correspond to harmomic
%     mean when the mimetic innerproduct B is diagonal.
%
   
   % (a)
   invFaceTrans     = zeros(g.faces.num, 1);
   invFaceTrans (f) = t.*sum(g.faces.neighbors(f,:) ~= 0, 2);
   invFaceTrans (f) = 1./invFaceTrans (f);
   
   % (b)
   numNodes        = diff(g.faces.nodePos); 
   invSubFaceTrans = invFaceTrans(fno).*double(numNodes(fno));

   % (c)
   B               = B + spdiags(invSubFaceTrans, 0, size(B, 1), size(B, 2));
end



%% Matlab code calling mex function invertSmallMatrices.c
function iA = invertDiagonalBlocks(A, sz)
   sz = int32(sz);
   blocks = matrixBlocksFromSparse(A, sz);

   [i, j] = blockDiagIndex(sz, sz);
   iA  = sparse(i,j, invv(blocks, sz));
end



%{
%% Pure Matlab code using inv
function iA = invertDiagonalBlocks(A, sz)
   V = zeros([sum(sz .^ 2), 1]);
   [p1, p2] = deal(0);

   for b = 1 : numel(sz),
      n  = sz(b);
      n2 = n * n;
      i  = p1 + (1 : n);

      V(p2 + (1 : n2)) = inv(full(A(i, i)));

      p1 = p1 + n;
      p2 = p2 + n2;
   end

   [I, J] = blockDiagIndex(sz);
   iA     = sparse(I, J, V);
end
%}
