function [R, N, nmax, cyclefree] = findingCycles(G, flux)
% Checks if there are cycles in the flux-field.
% SYNOPSIS:
%   [R, N, nmax, cyclefree] = findingCycles(G, flux)
%
% REQUIRED PARAMETERS:
%   G       - Grid data structure as described by grid_structure.
%   flux    - Vector of fluxes where element i represent the flux through
%             face i. 
%
% RETURNS:
%   R           - Ratio of cells in cycles. 
%   N           - The number of cycles
%   nmax        - The size of the largest cycle
%   cyclefree   - Logical parameter, true if there are no cycles and false
%                 if there are cycles.

%{
Copyright 2012 Christine M. Ø. Haugland.

This file is part of the Hybrid Mimetic and Multipoint module(HybridMethods).
HybridMethods is a module for the MATLAB Reservoir Simulation Toolbox (MRST).

HybridMethods is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

HybridMethods is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HybridMethods.  If not, see <http://www.gnu.org/licenses/>.

%} 


locN = double(G.faces.neighbors);


% The fluxes are placed in the matrix fluxcopy. The value is placed in 
% fluxcopy(i,j) if the flux from cell i to cell j is positive. If the flux
% is negative, we change the sign of the flux and place it in fluxcopy(j,i)  

fluxcopy = flux;
for i = 1:length(locN)
    if(locN(i,1) == 0)
        locN(i,1) = locN(i,2);
      %  fluxcopy(i) = 1;
    elseif(locN(i,2) == 0)
        locN(i,2) = locN(i,1);
      %  fluxcopy(i) = 1;
    end
end

for i = 1:length(fluxcopy)
    % Setting very small fluxes to zero.
    if(abs(fluxcopy(i))<=max(abs(flux))*1.0e-13)
        fluxcopy(i) = 0;
    end
    if(fluxcopy(i)<0)
        temp = locN(i,1);
        locN(i,1) = locN(i,2);
        locN(i,2) = temp;
        newtemp = fluxcopy(i);
        fluxcopy(i) = -1*newtemp;
    end
end

adj = sparse(locN(:,1), locN(:,2), fluxcopy);

% Place one on the diagonal such that dmperm works correctly. 
for i = 1:G.cells.num
    adj(i,i) = 1;
end

[p, q, r] = dmperm(adj);
if(length(r) == G.cells.num+1)
    cyclefree = true;
else
    cyclefree = false;
end

t = length(r)-1;
r = r';


u = diff(r)-1;
N = length(find(u ~= 0));
R = 1 - t/G.cells.num;
nmax = max(u); 
end